//#define TIMING
//#define VERBOSE
/**
 * @file yac_ioserver.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include "fields.h"
#include "yac_interface.h"
#include "generate_reg2d.h"
#include "utils.h"
#include "config_xml.h"

char * fieldName;

void set_grid_size ( int truncation, int * NUM_CELLS_X, int * NUM_CELLS_Y );

void get_dimensions ( int block_size[2], int block_pos[2], int cyclic[2],
                      int nbr_points[2], int nbr_cells[2], int nbr_procs[2],
                      int rank, int NUM_CELLS_X, int NUM_CELLS_Y );

void set_connectivity ( int comp_id, int grid_id, int block_size[2],
                        int block_pos[2], int nbr_procs[2], int rank, int cyclic[2], int nbr_points[2],
                        int nbr_cells[2], int NUM_CELLS_X, int NUM_CELLS_Y );

int set_points_and_elements ( int comp_id, int block_size[2], int block_pos[2], int cyclic[2],
                              int nbr_points[2], int nbr_cells[2], int NUM_CELLS_X, int NUM_CELLS_Y, int * grid_id );

int set_fields ( int point_id );

/* ===================================================== */


int main () {

  /* -------------- Initialisation of MPI -------------- */

  MPI_Init (0, NULL);

  /* Initialisation of YAC includes the reading of the XML file */

#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

#ifdef TIMING
  double tic, toc, time;
#endif

  yac_cinit ("coupling.xml", "coupling.xsd");

  /* -------------- Component definition -------------- */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "IOSERVER";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "yac_ioserver", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

#ifdef TIMING
  MPI_Barrier(local_comm);
  tic=MPI_Wtime();
#endif

  int ierror = 0;
  int rank, size;
  int NUM_CELLS_X;
  int NUM_CELLS_Y;

  int point_id;
  
  int grid_id;
  int nbr_fields;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  /* -------------- Grid definition -------------- 

     supported : ECHAM T21, T31, T42, T63, T85, T106, T127, T159, and T255

     default: 256 x 128

  */

  set_grid_size (21, &NUM_CELLS_X, &NUM_CELLS_Y );

  int total_nbr_cells[2] = {NUM_CELLS_X, NUM_CELLS_Y};
  int nbr_procs[2];

  generate_reg2d_decomp(total_nbr_cells, size, nbr_procs);

  int block_pos[2];
  int block_size[2];
  int nbr_cells[2];
  int cyclic[2];
  int nbr_points[2];

  get_dimensions ( block_size, block_pos, cyclic, nbr_points, nbr_cells, nbr_procs, rank, NUM_CELLS_X, NUM_CELLS_Y );

  point_id = set_points_and_elements ( comp_id, block_size, block_pos, cyclic,
                                       nbr_points, nbr_cells, NUM_CELLS_X, NUM_CELLS_Y, &grid_id );

  set_connectivity ( comp_id, grid_id, block_size, block_pos, nbr_procs,
                     rank, cyclic, nbr_points, nbr_cells, NUM_CELLS_X, NUM_CELLS_Y );

  /* -------------- Field definition -------------- */

  nbr_fields = set_fields ( point_id );

  int * field_ids  = xmalloc ( nbr_fields * sizeof(field_ids) );

  yac_cget_field_ids ( nbr_fields, field_ids );

#ifdef TIMING
  toc=MPI_Wtime();
  time = toc-tic;
  printf ("yac_ioserver: Time for initialisation %f\n", time );
#endif

  /* Search */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

#ifdef TIMING
  MPI_Barrier(local_comm);
  tic=MPI_Wtime();
#endif

  yac_csearch ( &ierror );

#ifdef TIMING
  toc=MPI_Wtime();
  time = toc-tic;
  printf ("yac_ioserver: Time for search %f\n", time );
#endif

  /* --------------  Time loop -------------- */

  int num_steps = 10;

  for (int step = 0; step < num_steps; ++step) {

#ifdef TIMING
    MPI_Barrier(local_comm);    
    tic=MPI_Wtime();
#endif

    {
      unsigned nbr_hor_cells = nbr_cells[0] * nbr_cells[1];

      int err;
      int info;

      double * recv_data = xmalloc(nbr_hor_cells * sizeof(*recv_data));

      for (int i = 0; i < nbr_hor_cells; ++i)
        recv_data[i] = -10;

      double *collection_data[1];

      for ( int i = 0; i < nbr_fields; ++i ) {

        collection_data[0] = recv_data;
        yac_cget(field_ids[i], 1, collection_data, &info, &err);

        if (info > 0 && info < 7 ) {
          yac_cget_field_name_from_id ( field_ids[i], fieldName );
          printf ( "ioserver has something to write out for %s\n", fieldName);
        }
      }
      free(recv_data);
    }

#ifdef TIMING
    toc=MPI_Wtime();
    time = toc-tic;
    printf ("yac_ioserver: Time for pong %f\n", time );
#endif

  }

  /* -------------- Finalising and cleanup -------------- */

  yac_cfinalize();

  MPI_Finalize();

  free(field_ids);

  return EXIT_SUCCESS;

}

/* ----------------------------------------------------------------------- */

void set_connectivity ( int comp_id,
                        int grid_id,
                        int block_size[2],
                        int block_pos[2],
                        int nbr_procs[2],
                        int rank,
                        int cyclic[2],
                        int nbr_points[2],
                        int nbr_cells[2],
                        int NUM_CELLS_X,
                        int NUM_CELLS_Y ) {

  int * global_global_cell_id = xmalloc(NUM_CELLS_Y * (NUM_CELLS_X + 2) * sizeof(*global_global_cell_id));
  int * global_global_cell_id_rank = xmalloc(NUM_CELLS_Y * (NUM_CELLS_X + 2) * sizeof(*global_global_cell_id_rank));
  int * global_global_corner_id = xmalloc((NUM_CELLS_Y + 1) * (NUM_CELLS_X + 3) * sizeof(*global_global_corner_id));

  for (int j = 0, id = 0; j < NUM_CELLS_Y; ++j) {

    for (int i = 1; i <= NUM_CELLS_X; ++i) {

      global_global_cell_id[j * (NUM_CELLS_X + 2) + i] = id;
      global_global_cell_id_rank[j * (NUM_CELLS_X + 2) + i] =
        (id%NUM_CELLS_X) / block_size[0] + ((id/NUM_CELLS_X) / block_size[1]) * nbr_procs[0];
      id++;
    }
  }
  for (int i = 0; i < NUM_CELLS_Y; ++i) {
    global_global_cell_id[i * (NUM_CELLS_X + 2) + 0] = global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X];
    global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + 0] = global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + NUM_CELLS_X];
    global_global_cell_id[i * (NUM_CELLS_X + 2) + NUM_CELLS_X+1] = global_global_cell_id[i * (NUM_CELLS_X + 2) + 1];
    global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + NUM_CELLS_X+1] = global_global_cell_id_rank[i * (NUM_CELLS_X + 2) + 1];
  }

  if (nbr_procs[0] == 1) {

    for (int j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j) {

      for (int i = 1; i <= NUM_CELLS_X; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + i] = id++;
    }

  } else {

    for (int j = 0, id = 0; j < NUM_CELLS_Y + 1; ++j) {

      for (int i = 1; i <= NUM_CELLS_X + 1; ++i)
        global_global_corner_id[j * (NUM_CELLS_X + 3) + i] = id++;
    }

    for (int i = 0; i < NUM_CELLS_Y + 1; ++i) {

      global_global_corner_id[i * (NUM_CELLS_X + 3) + 0] = global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 1];
      global_global_corner_id[i * (NUM_CELLS_X + 3) + NUM_CELLS_X + 2] = global_global_corner_id[i * (NUM_CELLS_X + 3) + 1];
    }
  }

  for (int i = 0; i < (NUM_CELLS_Y)*(NUM_CELLS_X + 2); ++i)
    if(global_global_cell_id_rank[i] == rank) global_global_cell_id_rank[i] = -1;

  int * local_global_cell_id = xmalloc(nbr_cells[1] * nbr_cells[0] * sizeof(*local_global_cell_id));
  int * local_global_cell_id_rank = xmalloc(nbr_cells[1] * nbr_cells[0] * sizeof(*local_global_cell_id_rank));
  int * local_global_corner_id = xmalloc(nbr_points[1] * nbr_points[0] * sizeof(*local_global_corner_id));
  int * local_global_corner_id_rank = xmalloc(nbr_points[1] * nbr_points[0] * sizeof(*local_global_corner_id_rank));

  int offset_x, offset_y;

  if (cyclic[0])
    offset_x = 1;
  else
    offset_x = block_size[0] * block_pos[0];
  offset_y = block_size[1] * block_pos[1] - (block_pos[1] != 0);

  for (int j = 0; j < nbr_cells[1]; ++j) {
    for (int i = 0; i < nbr_cells[0]; ++i) {
      local_global_cell_id[j * nbr_cells[0] + i] = global_global_cell_id[(j+offset_y) * (NUM_CELLS_X + 2) + i+offset_x];
      local_global_cell_id_rank[j * nbr_cells[0] + i] = global_global_cell_id_rank[(j+offset_y) * (NUM_CELLS_X + 2) +i+offset_x];
    }
  }

  free(global_global_cell_id_rank);
  free(global_global_cell_id);

  for (int j = 0; j < nbr_points[1]; ++j)
    for (int i = 0; i < nbr_points[0]; ++i)
      local_global_corner_id[j * nbr_points[0] + i] = global_global_corner_id[(j+offset_y) * (NUM_CELLS_X + 3) +i+offset_x];

  free(global_global_corner_id);

  for (int j = 0; j < nbr_points[1]; ++j)
    for (int i = 0; i < nbr_points[0]; ++i)
      local_global_corner_id_rank[j * nbr_points[0] + i] = -1;

  for (int i = 0; i < nbr_points[0]-2; ++i) {
    local_global_corner_id_rank[0 * nbr_points[0] + i] =
            local_global_cell_id_rank[0 * nbr_cells[0] + i];
    local_global_corner_id_rank[(nbr_points[1]-1) * nbr_points[0] + i] =
            local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + i];
  }
  for (int j = 0; j < nbr_points[1]-2; ++j) {
    local_global_corner_id_rank[j * nbr_points[0] + 0] =
            local_global_cell_id_rank[j * nbr_cells[0] + 0];
    local_global_corner_id_rank[j * nbr_points[0] + nbr_points[0]-1] =
            local_global_cell_id_rank[j * nbr_cells[0] + nbr_cells[0]-1];
  }

  local_global_corner_id_rank[(nbr_points[1]-1) * nbr_points[0] + nbr_points[0]-1] =
            local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + nbr_cells[0] - 1];

  local_global_corner_id_rank[(nbr_points[1]-2) * nbr_points[0] + 0] =
            local_global_cell_id_rank[(nbr_cells[1]-2) * nbr_cells[0] + 0];

  local_global_corner_id_rank[0 * nbr_points[0] + nbr_points[0] - 2] =
            local_global_cell_id_rank[0 * nbr_cells[0] + nbr_cells[0] - 2];

  local_global_corner_id_rank[(nbr_points[1]-2) * nbr_points[0] + nbr_points[0]-1] =
            local_global_cell_id_rank[(nbr_cells[1]-2) * nbr_cells[0] + nbr_cells[0] - 1];

  local_global_corner_id_rank[(nbr_points[1]-1) * nbr_points[0] + nbr_points[0]-2] =
            local_global_cell_id_rank[(nbr_cells[1]-1) * nbr_cells[0] + nbr_cells[0] - 2];

  int * cell_core_mask =
    xmalloc(nbr_cells[1] * nbr_cells[0] * sizeof(*cell_core_mask));
  int * corner_core_mask =
    xmalloc(nbr_points[1] * nbr_points[0] * sizeof(*corner_core_mask));

  for (int i = 0; i < nbr_cells[1]*nbr_cells[0]; ++i)
    cell_core_mask[i] = local_global_cell_id_rank[i] == -1;
  for (int i = 0; i < nbr_points[1]*nbr_points[0]; ++i)
    corner_core_mask[i] = local_global_corner_id_rank[i] == -1;

  yac_cset_global_index(local_global_cell_id, CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, CELL, grid_id);
  yac_cset_global_index(local_global_corner_id, CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, CORNER, grid_id);

  free(corner_core_mask);
  free(cell_core_mask);
  free(local_global_corner_id_rank);
  free(local_global_corner_id);
  free(local_global_cell_id_rank);
  free(local_global_cell_id);
}

/* ----------------------------------------------------------------------- */

int set_points_and_elements ( int comp_id,
                              int block_size[2],
                              int block_pos[2],
                              int cyclic[2],
                              int nbr_points[2],
                              int nbr_cells[2],
                              int NUM_CELLS_X,
                              int NUM_CELLS_Y,
                              int * grid_id ) {

  double * global_x_vertices = xmalloc((NUM_CELLS_X+3) * sizeof(*global_x_vertices));
  double * global_y_vertices = xmalloc((NUM_CELLS_Y+1) * sizeof(*global_y_vertices));;
  double * x_vertices;
  double * y_vertices;
  double * x_points;
  double * y_points;

  int point_id;

  for (int i = 0; i < NUM_CELLS_X+3; ++i)
    global_x_vertices[i] = 0.0 + (2.0 * M_PI / ((double)NUM_CELLS_X)) * (double)(i-1);
  for (int i = 0; i < NUM_CELLS_Y; ++i)
    global_y_vertices[i] = -M_PI_2 + (M_PI / ((double)NUM_CELLS_Y)) * (double)i;
  global_y_vertices[NUM_CELLS_Y] = M_PI_2;

  if (cyclic[0])
    x_vertices = x_points = global_x_vertices + 1;
  else
    x_vertices = x_points = global_x_vertices + block_size[0] * block_pos[0];

  y_vertices = y_points = global_y_vertices + block_size[1] * block_pos[1];

  if (block_pos[1] != 0) {
    y_vertices -= 1;
    y_points -= 1;
  }

  yac_cdef_grid_reg2d ( "grid1", nbr_points, cyclic, x_vertices, y_vertices, grid_id);

  yac_cdef_points_reg2d ( *grid_id, nbr_cells, CELL, x_points, y_points, &point_id );

  free(global_x_vertices);
  free(global_y_vertices);

  return point_id;

}

/* ----------------------------------------------------------------------- */

int set_fields ( int point_id ) {

  struct yac_couple_config * couple_config =
    yac_couple_config_xml_new ("coupling.xml", "coupling.xsd");

  int comp_id =
    yac_couple_config_get_component_id_by_name(couple_config, "IOSERVER");
  size_t num_couples = yac_couple_config_get_num_couples(couple_config);

  int field_count = 0;

  for (size_t couple_idx = 0; couple_idx < num_couples; ++couple_idx) {

    int couple_comp_ids[2];
    yac_couple_config_get_couple_component_ids(
      couple_config, couple_idx, couple_comp_ids);

    if ((couple_comp_ids[0] != comp_id) && (couple_comp_ids[1] != comp_id))
      continue;

    size_t num_couple_transients =
      yac_couple_config_get_num_couple_transients(couple_config, couple_idx);

    for (size_t couple_transient_idx = 0;
         couple_transient_idx < num_couple_transients; ++couple_transient_idx) {

      int source_comp_id, target_comp_id;
      yac_couple_config_get_transient_couple_component_ids(
        couple_config, couple_idx, couple_transient_idx,
        &source_comp_id, &target_comp_id);
      char const * transient_name =
        yac_couple_config_get_transient_name(
          couple_config, couple_idx, couple_transient_idx);

      if (source_comp_id == comp_id) {
        fprintf(
          stderr, "ERROR: IOSERVER is set as source for transient \"%s\"\n",
          transient_name);
        exit(EXIT_FAILURE);
      } else if (comp_id == target_comp_id) {

        field_count++;
        int dummy;
        yac_cdef_field(transient_name, comp_id, &point_id, 1, &dummy);
        fprintf(
          stderr, "defined field with standard name %s\n", transient_name);
      } else {
        fprintf(
          stderr, "ERROR: IOSERVER is neither source nor target for \"%s\", "
                  "this should not have happend\n", transient_name);
        exit(EXIT_FAILURE);
      }
    }
  }
  yac_couple_config_delete(couple_config);
  return field_count;
}

/* ----------------------------------------------------------------------- */

void set_grid_size ( int truncation,
                     int * NUM_CELLS_X,
                     int * NUM_CELLS_Y ) {

  switch (truncation) {

  case (21):

    {
      *NUM_CELLS_X = 64;
      *NUM_CELLS_Y = 32;
      break;
    }

  case (31):

    {
      *NUM_CELLS_X = 96;
      *NUM_CELLS_Y = 48;
      break;
    }

  case (42):

    {
      *NUM_CELLS_X = 128;
      *NUM_CELLS_Y = 64;
      break;
    }

  case (63):

    {
      *NUM_CELLS_X = 192;
      *NUM_CELLS_Y = 96;
      break;
    }

  case (85):

    {
      *NUM_CELLS_X = 256;
      *NUM_CELLS_Y = 128;
      break;
    }

  case (106):

    {
      *NUM_CELLS_X = 320;
      *NUM_CELLS_Y = 160;
      break;
    }

  case (127):

    {
      *NUM_CELLS_X = 384;
      *NUM_CELLS_Y = 192;
      break;
    }

  case (159):

    {
      *NUM_CELLS_X = 480;
      *NUM_CELLS_Y = 240;
      break;
    }

  case (255):

    {
      *NUM_CELLS_X = 768;
      *NUM_CELLS_Y = 384;
      break;
    }

  default:

    {
      *NUM_CELLS_X = 256;
      *NUM_CELLS_Y = 128;
      break;
    }

  };

}

/* ----------------------------------------------------------------------- */

void get_dimensions ( int block_size[2],
                      int block_pos[2],
                      int cyclic[2],
                      int nbr_points[2],
                      int nbr_cells[2],
                      int nbr_procs[2],
                      int rank,
                      int NUM_CELLS_X,
                      int NUM_CELLS_Y ) {

  block_pos[0] = rank%nbr_procs[0];
  block_pos[1] = rank/nbr_procs[0];
  block_size[0] = (NUM_CELLS_X + nbr_procs[0] - 1)/nbr_procs[0];
  block_size[1] = (NUM_CELLS_Y + nbr_procs[1] - 1)/nbr_procs[1];
  nbr_cells[0] = MIN(block_size[0], NUM_CELLS_X - block_size[0] * block_pos[0]);
  nbr_cells[1] = MIN(block_size[1], NUM_CELLS_Y - block_size[1] * block_pos[1]);
  cyclic[0] = nbr_procs[0] == 1;
  cyclic[1] = 0;

  // halo
  if (!cyclic[0])
    nbr_cells[0] += 2;
  if (nbr_procs[1] > 1) {

    nbr_cells[1] += 1;

    if ((block_pos[1] != 0) && (block_pos[1] != nbr_procs[1]-1))
      nbr_cells[1] += 1;
  }

  if (cyclic[0])
    nbr_points[0] = nbr_cells[0];
  else
    nbr_points[0] = nbr_cells[0]+1;

  nbr_points[1] = nbr_cells[1] + 1;

}

/* ----------------------------------------------------------------------- */

