package de.mpg.mpimet.yacgui.model;

import de.mpg.mpimet.yacgui.model.Model.Align;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Helper class to encapsulate the coordinates of couplings
 * 
 * @author Maxim Yastremsky
 */
public class Connection {
	
	private Align sourceAlign;
	private Align targetAlign;
	private int sourceIndex;
	private int targetIndex;
	private boolean hasInterpolations;
	
	/**
	 * Creates a new Connection.
	 * 
	 * @param sourceAlign	LEFT or RIGHT
	 * @param sourceIndex	index of source transient panel
	 * @param targetIndex	index of target transient panel
	 * @param hasInterpolation switch
	 */
	public Connection(Align sourceAlign, int sourceIndex, int targetIndex, boolean hasInterpolation) {
		this(sourceAlign, sourceAlign.getOpposite(), sourceIndex, targetIndex, hasInterpolation);
	}
	
	/**
	 * Creates a new Connection.
	 * 
	 * @param sourceAlign
	 *            LEFT or RIGHT
	 * @param targetAlign
	 *            LEFT or RIGHT (other than sourceAlign)
	 * @param sourceIndex
	 *            index of source transient panel
	 * @param targetIndex
	 *            index of target transient panel
	 * @param hasInterpolations switch
	 * @throws IllegalArgumentException
	 *             if targetAlign is not different from sourceAlign
	 */
	public Connection(Align sourceAlign, Align targetAlign, int sourceIndex, int targetIndex, boolean hasInterpolations) {
		this.sourceAlign = sourceAlign;
		if (targetAlign != sourceAlign.getOpposite()) {
			throw new IllegalArgumentException("Parameter Align targetAlign has to be other than sourceAlign.");
		}
		this.targetAlign = targetAlign;
		this.sourceIndex = sourceIndex;
		this.targetIndex = targetIndex;
		this.hasInterpolations = hasInterpolations;
	}
	
	/**
	 * @return the sourceAlign
	 */
	public Align getSourceAlign() {
		return sourceAlign;
	}
	
	/**
	 * @return the targetAlign
	 */
	public Align getTargetAlign() {
		return targetAlign;
	}
	
	/**
	 * @return the sourceIndex
	 */
	public int getSourceIndex() {
		return sourceIndex;
	}
	
	/**
	 * @return the targetIndex
	 */
	public int getTargetIndex() {
		return targetIndex;
	}
	
	public boolean hasInterpolations() {
		return hasInterpolations;
	}
	
	public void setHavingInterpolations(boolean hasInterp) {
		hasInterpolations = hasInterp;
	}
	
	/**
	 * Gets the index at given Align.
	 * 
	 * @param align	LEFT or RIGHT
	 * @return index at given align
	 */
	public int getIndexAt(Align align) {
		if (align == sourceAlign) {
			return sourceIndex;
		}
		return targetIndex;
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (hasInterpolations ? 1231 : 1237);
		result = prime * result
				+ ((sourceAlign == null) ? 0 : sourceAlign.hashCode());
		result = prime * result + sourceIndex;
		result = prime * result
				+ ((targetAlign == null) ? 0 : targetAlign.hashCode());
		result = prime * result + targetIndex;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Connection other = (Connection) obj;
		if (hasInterpolations != other.hasInterpolations)
			return false;
		if (sourceAlign != other.sourceAlign)
			return false;
		if (sourceIndex != other.sourceIndex)
			return false;
		if (targetAlign != other.targetAlign)
			return false;
		if (targetIndex != other.targetIndex)
			return false;
		return true;
	}

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer("[");
		buf.append(sourceAlign).append(":");
		buf.append(sourceIndex).append("]->[");
		buf.append(targetAlign).append(":");
		buf.append(targetIndex).append("]");
		return buf.toString();
	}

}
