/**
 * @file yac_interface.h
 * @brief C User Interface
 *
 * An overview of the C User Interface can be found \ref cinterface "here".
 * For a detailed description of the Fortran interface please see
 * \ref mo_yac_finterface.f90 and \ref yac_finterface.f90
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 * @par Warranty
 *
 * This code has been tested up to a certain level. Defects and weaknesses,
 * which may be included in the code, do not establish any warranties by the
 * authors. The authors do not make any warranty, express or implied, or assume any
 * liability or responsibility for the use, acquisition or application of this
 * software.
 *
 * @par C User API
 *
 * For a detailed description of the C interface please see \ref cinterface.
 * Information on the Fortran interface can be found here \ref finterface.
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef YAC_INTERFACE_H
#define YAC_INTERFACE_H

#include <mpi.h>

extern int const YAC_LOCATION_CELL;
extern int const YAC_LOCATION_CORNER;
extern int const YAC_LOCATION_EDGE;
#define YAC_MAX_CHARLEN (132)

/** \example test_dummy_coupling_c.c
 * This example simulates a whole model setup with three components (ocean,
 * atmosphere, io). It uses one process for each component.
 */

/** \example test_dummy_coupling2_c.c
 * This example simulates a whole model setup with two components (ocean,
 * atmosphere) where rank 2 is shared by both components.
 */

/** \example test_dummy_coupling3_c.c
 * This tests checks the support for multiple parallel YAC instances.
 * There are three main components and one dummy component. The dummy component
 * does not take part in any coupling. YAC instances are used.
 */

/** \example test_restart.c
 * This example simulates a restart of a coupled configuration with a grid redefiniton.
 */

/** \example test_restart2.c
 * This example simulates a bit more complex restart of a coupled configuration with a grid redefiniton.
 */

/* -----------------------------------------------------------------------------

    C and Fortran description of the user API

   ----------------------------------------------------------------------------- */

/** Elementary initialisation of the whole system

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @remark this call initialises the default YAC instance
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
*/
    void yac_cinit ( const char * xml_filename,
                     const char * schema_filename );

/** Elementary initialisation of the whole system

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
*/
    void yac_cinit_instance ( const char * xml_filename,
                              const char * schema_filename,
                              int * yac_instance_id );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @param[in] comm            MPI communicator
     @remark this call initialises the default YAC instance
     @remark the MPI communicator provided to this routine has to contain
             all processes that will take part in the coupling
     @remark this call is collective for all processes in comm
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
*/
    void yac_cinit_comm ( const char * xml_filename,
                          const char * schema_filename,
                          MPI_Comm comm );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[in]  comm            MPI communicator
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
*/
    void yac_cinit_comm_instance ( const char * xml_filename,
                                   const char * schema_filename,
                                   MPI_Comm comm, int * yac_instance_id );

/* -------------------------------------------------------------------------------- */

/** Dummy for initialisation of the whole system
     @remark This routine can be called instead of \ref yac_cinit or
             \ref yac_cinit_instance, if the local process does not wish to
             initialise a YAC instance.
*/
    void yac_cinit_dummy();

/** Dummy initialisation of the whole system using a user-provided
 *  MPI world communicator

     @param[in] comm MPI communicator
     @remark This routine can be called instead of \ref yac_cinit_comm or
             \ref yac_cinit_comm_instance, if the local process does not wish to
             initialise a YAC instance.
*/
    void yac_cinit_comm_dummy ( MPI_Comm comm );

/* -------------------------------------------------------------------------------- */

/** Clean-up YAC (same as yac_cfinalize but without MPI_Finalize; see \ref CRestart)
     @remark In case the user initialised yaxt himself, he has to finalise it after
             yac_ccleanup in order to be able to call yac_cinit again.
     @remark cleans up default YAC instance
*/
     void yac_ccleanup ();

/** Clean-up YAC (same as yac_cfinalize_instance but without MPI_Finalize;
 *  see \ref CRestart)
     @param[in] yac_instance_id id of the YAC instance to be cleaned up
     @remark In case the user initialised yaxt himself, he has to finalise it after
             yac_ccleanup in order to be able to call yac_cinit again.
*/
     void yac_ccleanup_instance ( int yac_instance_id );

/* -------------------------------------------------------------------------------- */

/** Cleans up default YAC instance (if available) and finales MPI,
 *  if it was initialised by YAC
*/
     void yac_cfinalize ();

/** Cleans up provided YAC instance and finales MPI,
 *  if it was initialised by YAC
*/
     void yac_cfinalize_instance ();

/* -------------------------------------------------------------------------------- */

/** Definition of job start and end datetime for the default YAC instance

     @param[in] start_datetime calendar job start datetime
     @param[in] end_datetime   calendar job end datetime
 */

     void yac_cdef_datetime ( const char * start_datetime,
                              const char * end_datetime );

/** Definition of job start and end datetime

     @param[in] yac_instance_id id of the YAC instance
     @param[in] start_datetime  calendar job start datetime
     @param[in] end_datetime    calendar job end datetime
 */

     void yac_cdef_datetime_instance ( int yac_instance_id,
                                       const char * start_datetime,
                                       const char * end_datetime );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the component for the default YAC instance

     @param[in] comp_name name of the component
     @param[in] comp_id   component Id
     @remark this call is collective for all processes that initialised the
             default YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comp ( const char * comp_name,
                          int * comp_id );

/** Elementary definition of the component.

     @param[in] yac_instance_id id of the YAC instance
     @param[in] comp_name name of the component
     @param[in] comp_id   component Id
     @remark this call is collective for all processes that initialised the
             provided YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comp_instance ( int yac_instance_id,
                                   const char * comp_name,
                                   int * comp_id );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the components for the default YAC instance

     @param[in] comp_names names of the components
     @param[in] num_comps  number of components
     @param[in] comp_ids   component Id's
     @remark this call is collective for all processes that initialised the
             default YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comps ( const char ** comp_names,
                           int num_comps,
                           int * comp_ids );

/** Elementary definition of the components.

     @param[in] yac_instance_id id of the YAC instance
     @param[in] comp_names names of the components
     @param[in] num_comps  number of components
     @param[in] comp_ids   component Id's
     @remark this call is collective for all processes that initialised the
             provided YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comps_instance ( int yac_instance_id,
                                    const char ** comp_names,
                                    int num_comps,
                                    int * comp_ids );

/* -------------------------------------------------------------------------------- */

/** Setting io_proc_step for a component.

     @param[in] comp_id      - component Id
     @param[in] io_proc_step - stride between component processes doing io
                               (e.g. writing weight files)
*/

     void yac_cset_comp_io_proc_step ( int comp_id,
                                       int io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Getting io_proc_step from a component.

     @param[in] comp_id       - component Id
     @param[out] io_proc_step - stride between component processes doing io
                                (e.g. writing weight files)
*/

     void yac_cget_comp_io_proc_step ( int comp_id,
                                       int * io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Providing the local MPI communicator

     @param[in] comp_id            - local component ID (from yac_cdef_comp), handle to the component struct
     @param[in] local_communicator - returned local communicator

*/
     void yac_cget_localcomm ( MPI_Comm *local_communicator,
                               int comp_id );

/* -------------------------------------------------------------------------------- */

/** Providing a MPI communicator for communication between the component
 *  root processes

     @param[in] local_communicator returned local communicator
     @param[in] comp_name1         name of the first component
     @param[in] comp_name2         name of the second component
     @remark the components have to defined in the default YAC instance
*/
     void yac_cget_pair_rootcomm ( MPI_Comm *local_communicator,
                                   const char * comp_name1,
                                   const char * comp_name2 );

/** Providing a MPI communicator for communication between the component
 *  root processes

     @param[in] yac_instance_id    id of the YAC instance
     @param[in] local_communicator returned local communicator
     @param[in] comp_name1         name of the first component
     @param[in] comp_name2         name of the second component
*/
     void yac_cget_pair_rootcomm_instance ( int yac_instance_id,
                                            MPI_Comm *local_communicator,
                                            const char * comp_name1,
                                            const char * comp_name2 );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D regular grids.

     @param[in] grid_id    handle to the grid struct (returned from yac_cdef_grid)
     @param[in] nbr_points number of points
     @param[in] location   location of points [edge|vertex|center]
     @param[in] x_points   array of point longitudes
     @param[in] y_points   array of point latitudes
     @param[in] point_id   returned handle to the points struct
*/

     void yac_cdef_points_reg2d ( int const grid_id,
                                  int const *nbr_points,
                                  int const location,
                                  double const *x_points,
                                  double const *y_points,
                                  int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D unstructured grids.

     @param[in] grid_id    handle to the grid struct (returned from yac_cdef_grid)
     @param[in] nbr_points number of points
     @param[in] location   location of points [edge|vertex|center]
     @param[in] x_points   array of point longitudes
     @param[in] y_points   array of point latitudes
     @param[in] point_id   returned handle to the points struct
*/

     void yac_cdef_points_unstruct ( int const grid_id,
                                     int const nbr_points,
                                     int const location,
                                     double const *x_points,
                                     double const *y_points,
                                     int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a 2d regular grid

     @param[in]  grid_name    name of the grid
     @param[in]  nbr_vertices 2d array containing the number of vertices in each dimension
     @param[in]  cyclic       2d array containing information about cyclic behaviour in each dimension
     @param[in]  x_vertices   array of vertex longitudes
     @param[in]  y_vertices   array of vertex latitudes
     @param[out] grid_id      id of generated grid
     @remark on the local order of edges:
             generate for each edge a tuple containing the local indices of the
             adjacent vertices (lowest index first) and sort this list by the
             vertex indices; this gives the the local order of edges
*/
     void yac_cdef_grid_reg2d ( const char * grid_name,
                                int nbr_vertices[2],
                                int cyclic[2],
                                double *x_vertices,
                                double *y_vertices,
                                int *grid_id);

/* -------------------------------------------------------------------------------- */

/** Definition of an unstructured grid

     @param[in]  grid_name             name of the grid
     @param[in]  nbr_vertices          number of vertices
     @param[in]  nbr_cells             number of cells
     @param[in]  num_vertices_per_cell array containing the number of vertices for each cell
     @param[in]  x_vertices            array of vertex longitudes
     @param[in]  y_vertices            array of vertex latitudes
     @param[in]  cell_to_vertex        connectivity of vertices belonging to cells\n
                                       (the vertex indices per cell have to be in clockwise or counterclockwise ordering)
     @param[out] grid_id               id of generated grid
*/
     void yac_cdef_grid_unstruct ( const char * grid_name,
                                   int nbr_vertices,
                                   int nbr_cells,
                                   int *num_vertices_per_cell,
                                   double *x_vertices,
                                   double *y_vertices,
                                   int *cell_to_vertex,
                                   int *grid_id);

/* -------------------------------------------------------------------------------- */

/** Set global ids for a grid

     @param[in] global_index array of global indices
     @param[in] location     cell/vertex/edge
     @param[in] grid_id      grid id
     @remark global indices are to be provided in the local order of their
             respective cells/vertices/edges
*/

     void yac_cset_global_index ( int const * global_index,
                                  int location,
                                  int grid_id);

/* -------------------------------------------------------------------------------- */

/** Set core mask for a grid

     @param[in] is_core   0 for cells/vertices/edges that are halos,
                          1 for cells/vertices/edges that are core
     @param[in] location  cell/vertex/edge
     @param[in] grid_id   grid id
     @remark cells/vertices/edges who are halos are not used or set
             in a put or get operation
     @remark core mask values are to be provided in the local order of their
             respective cells/vertices/edges
*/

     void yac_cset_core_mask ( int const * is_core,
                               int location,
                               int grid_id);

/* -------------------------------------------------------------------------------- */

/** Set the default mask for points

     @param[in] is_valid   0 for points that are masked out, 1 for valid points
     @param[in] points_id  id of points/cells
*/

     void yac_cset_mask ( int const * is_valid,
                          int points_id );

/* -------------------------------------------------------------------------------- */

/** define a mask for a grid

     @param[in] grid_id    grid ID
     @param[in] nbr_points number of points
     @param[in] location   cell/vertex/edge
     @param[in] is_valid   0 for points that are masked out, 1 for valid points
     @param[in] mask_id    returned handle to the mask struct
*/

     void yac_cdef_mask ( int const grid_id,
                          int const nbr_points,
                          int const location,
                          int const * is_valid,
                          int *mask_id );

/* -------------------------------------------------------------------------------- */

/** Definition of the coupling field (using default mask, if defined)

     @param[in] field_name               - character string providing the name of the coupling field
     @param[in] component_id             - component ID
     @param[in] point_ids                - point IDs
     @param[in] num_pointsets            - number of pointsets per grid
     @param[in] field_id                 - returned field_id which has to be used to identify coupling fields
                                       in yac_cput and yac_cget.
*/

       void yac_cdef_field ( char const * field_name,
                             int const component_id,
                             int const * point_ids,
                             int const num_pointsets,
                             int * field_id );

/* -------------------------------------------------------------------------------- */

/** Definition of the coupling field

     @param[in] field_name               - character string providing the name of the coupling field
     @param[in] component_id             - component ID
     @param[in] point_ids                - point IDs
     @param[in] mask_ids                 - mask IDs
     @param[in] num_pointsets            - number of pointsets per grid
     @param[in] field_id                 - returned field_id which has to be used to identify coupling fields
                                       in yac_cput and yac_cget.
*/

       void yac_cdef_field_mask ( char const * field_name,
                                  int const component_id,
                                  int const * point_ids,
                                  int const * mask_ids,
                                  int const num_pointsets,
                                  int * field_id );

/* -------------------------------------------------------------------------------- */

/** Receiving of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] recv_field         - receive buffer (all data is stored in one
                                                     contiguous part of the memory)
                                     dimensions: recv_field[collection_size][nbr_points]
     @param[in] info               - returned info argument indicating the action performed
     @param[in] ierror             - returned error
*/
     void yac_cget_ ( int const field_id,
                      int const collection_size,
                      double *recv_field,
                      int    *info,
                      int    *ierror );

/** Receiving of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] recv_field         - receive buffer
                                     dimensions: recv_field[collection_size][nbr_points]
     @param[in] info               - returned info argument indicating the action performed
     @param[in] ierror             - returned error
*/
     void yac_cget ( int const field_id,
                     int const collection_size,
                     double **recv_field,
                     int    *info,
                     int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer (all data is stored in one
                                                  contiguous part of the memory)
                                     dimensions send_field[collection_size]
                                                          [nbr_fields]
                                                          [nbr_points]
     @param[in] info               - returned info
     @param[in] ierror             - returned error
*/

     void yac_cput_ ( int const field_id,
                      int const collection_size,
                      double *send_field ,
                      int    *info,
                      int    *ierror );

/** Sending of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer
                                     dimensions send_field[collection_size]
                                                          [nbr_fields]
                                                          [nbr_points]
     @param[in] info               - returned info
     @param[in] ierror             - returned error
*/

     void yac_cput ( int const field_id,
                     int const collection_size,
                     double *** const send_field,
                     int *info,
                     int *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer
                                     dimensions send_field[collection_size * nbr_fields]
                                                          [nbr_points]
     @param[in] info               - returned info
     @param[in] ierror             - returned error
*/

     void yac_cput_ptr_ ( int const field_id,
                          int const collection_size,
                          double ** send_field,
                          int    *info,
                          int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Exchange of the coupling fields

     @param[in] send_field_id      -
     @param[in] recv_field_id      -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer (all data is stored in one
                                                  contiguous part of the memory)
                                     dimensions send_field[collection_size]
                                                          [nbr_fields]
                                                          [nbr_points]
     @param[in] recv_field         - receive buffer (all data is stored in one
                                                     contiguous part of the memory)
                                     dimensions: recv_field[collection_size][nbr_points]
     @param[in] send_info          - returned send info
     @param[in] recv_info          - returned recv info
     @param[in] ierror             - returned error
*/

     void yac_cexchange_ ( int const send_field_id,
                           int const recv_field_id,
                           int const collection_size,
                           double *send_field ,
                           double *recv_field,
                           int    *send_info,
                           int    *recv_info,
                           int    *ierror );


/** Exchange of the coupling fields

     @param[in] send_field_id      -
     @param[in] recv_field_id      -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer
                                     dimensions send_field[collection_size]
                                                          [nbr_fields]
                                                          [nbr_points]
     @param[in] recv_field         - receive buffer
                                     dimensions: recv_field[collection_size][nbr_points]
     @param[in] send_info          - returned send info
     @param[in] recv_info          - returned recv info
     @param[in] ierror             - returned error
*/

     void yac_cexchange ( int const send_field_id,
                          int const recv_field_id,
                          int const collection_size,
                          double *** const send_field,
                          double ** recv_field,
                          int *send_info,
                          int *recv_info,
                          int *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in] send_field_id      -
     @param[in] recv_field_id      -
     @param[in] collection_size    -
     @param[in] send_field         - send buffer
                                     dimensions send_field[collection_size * nbr_fields]
                                                          [nbr_points]
     @param[in] recv_field         - receive buffer
                                     dimensions: recv_field[collection_size][nbr_points]
     @param[in] send_info          - returned send info
     @param[in] recv_info          - returned recv info
     @param[in] ierror             - returned error
*/

     void yac_cexchange_ptr_ ( int const send_field_id,
                               int const recv_field_id,
                               int const collection_size,
                               double ** send_field,
                               double ** recv_field,
                               int    *send_info,
                               int    *recv_info,
                               int    *ierror );

/* -------------------------------------------------------------------------------- */

/** End of the definition phase for the default YAC instance,
 *  invocation of the search

     @param[out] ierror returned error
*/

void yac_csearch ( int * ierror );

/** End of the definition phase, invocation of the search

     @param[in]  yac_instance_id id of the YAC instance
     @param[out] ierror          returned error
*/

void yac_csearch_instance ( int yac_instance_id, int * ierror );

/* --------------------------------------------------------------------------------
           query routines
   -------------------------------------------------------------------------------- */

/** query routine for the start datetime of the job of the default YAC instance
*/

char * yac_cget_start_datetime ();

/** query routine for the start datetime of the job

     @param[in] yac_instance_id id of the YAC instance
*/

char * yac_cget_start_datetime_instance ( int yac_instance_id );

/** query routine for the end datetime of the job of the default YAC instance
*/

char * yac_cget_end_datetime ();

/** query routine for the end datetime of the job

     @param[in] yac_instance_id id of the YAC instance
*/

char * yac_cget_end_datetime_instance ( int yac_instance_id );

/** query routine for the yac version
*/

char * yac_cget_version ();

/* -------------------------------------------------------------------------------- */

/** query routine for number of coupling fields defined of the default YAC
 *  instance

     @param[in] nbr_fields         - number of fields defined on the calling process

*/

void yac_cget_nbr_fields ( int *nbr_fields );

/** query routine for number of coupling fields defined

     @param[in] yac_instance_id id of the YAC instance
     @param[in] nbr_fields      number of fields defined on the calling process

*/

void yac_cget_nbr_fields_instance ( int yac_instance_id, int *nbr_fields );

/* -------------------------------------------------------------------------------- */

/** query routine to get the list of field IDs for defined fields of the
 *  default YAC instance

     @param[in] nbr_fields number of fields defined on the calling process
     @param[in] field_ids  list of delivered field IDs

*/

void yac_cget_field_ids ( int nbr_fields, int *field_ids );

/** query routine to get the list of field IDs for defined fields

     @param[in] yac_instance_id id of the YAC instance
     @param[in] nbr_fields number of fields defined on the calling process
     @param[in] field_ids  list of delivered field IDs

*/

void yac_cget_field_ids_instance ( int yac_instance_id,
                                   int nbr_fields,
                                   int * field_ids );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] field_name field name that has been given to yac_cdef_field
     @param[in] field_id   ID as it has been provided by yac_cdef_field

*/

void yac_cget_field_id_from_name ( const char *field_name, int *field_id );

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name that has been given to yac_cdef_field
     @param[in] field_id        ID as it has been provided by yac_cdef_field

*/

void yac_cget_field_id_from_name_instance ( int yac_instance_id,
                                            const char *field_name,
                                            int *field_id );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] field_id   ID as provided by yac_cdef_field
     @param[in] field_name field name that has been given to yac_cdef_field

*/

void yac_cget_field_name_from_id ( int field_id, char *field_name );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID
 *  of the default YAC instance

     @param[in] field_name field name as provided by yac_cdef_field
     @param[in] role       NOTHING (0) || 1 SOURCE (1) || TARGET (2)
     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance

*/

void yac_cget_role_of_field_by_name ( char const * field_name, int *role );

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as provided by yac_cdef_field
     @param[in] role            NOTHING (0) || 1 SOURCE (1) || TARGET (2)
     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance

*/

void yac_cget_role_of_field_by_name_instance ( int yac_instance_id,
                                               char const * field_name,
                                               int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] field_id           - ID as provided by yac_cdef_field
     @param[in] role               - NOTHING (0) || 1 SOURCE (1) || TARGET (2)

*/

void yac_cget_role_of_field_by_id ( int field_id, int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field name of the default
 *  YAC instance

     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string );

/** query routine to get coupling interval for a given field name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_name_instance ( int yac_instance_id,
                                                const char * field_name,
                                                char *timestep_string );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field id

     @param[in] field_id        field ID (as representative for a field name)
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_id ( const int field_id,
                                     char *timestep_string );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field name of the default
 *  YAC instance

     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string model time step as iso8601 string

*/

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string );

/** query routine to get model time step for a given field name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string model time step as iso8601 string

*/

void yac_cget_model_timestep_name_instance ( int yac_instance_id,
                                             const char * field_name,
                                             char *timestep_string );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field id

     @param[in] field_id           - field ID (as representative for a field name)
     @param[in] timestep_string    - model time step as iso8601 string

*/

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string );

/* --------------------------------------------------------------------------------
           auxiliary routines
   -------------------------------------------------------------------------------- */

/**
 * redirect stderr and stdout into files with names depending on the process ID
 *
 * The original version was part of the OASIS4 coupler and has been adjusted to
 * work for yac.
 *
 * @param[in]  filestem   basename of stdout/stderr files
 * @param[in]  lenstr     significant length (without \0) of "filestem"
 * @param[in]  parallel   flag for direction of output
                             parallel = 1 : Create single output files for each process
                             otherwise    : Create output files only for process 0.
                                           The output of other processes is not redirected.
 * @param[in]  my_pe      local rank
 * @param[in]  npes       total number of processes in set
 */

void yac_redirstdout ( const char *filestem, int lenstr,
                       int parallel, int my_pe,
                       int npes );

/* -------------------------------------------------------------------------------- */

/** query routine to get the number of points in a grid

     @param[in] location
     @param[in] grid_id

*/

size_t yac_get_grid_size ( int location, int grid_id );

/* -------------------------------------------------------------------------------- */

/** query routine to get the number of points in a pointset

     @param[in] points_id

*/

size_t yac_get_points_size ( int points_id  );


#endif // YAC_INTERFACE_H

